@@ -77,7 +77,25 @@ static struct voltage_s {
   tl866ii_vcc_voltages[] = { { "3.3", 0x01 }, { "4", 0x02 },   { "4.5", 0x03 },
 			     { "5", 0x00 },   { "5.5", 0x04 }, { "6.5", 0x05 },
 			     { NULL, 0x00 } },
-  t48_vcc_voltages[] = { { "1.75", 0x01 },
+  t48_vcc_voltages[] = { { "3.3", 0x01 }, { "4", 0x02 },   { "4.5", 0x03 },
+			     { "5", 0x00 },   { "5.5", 0x04 }, { "6.5", 0x05 },
+			     { NULL, 0x00 } },
+  t48_vpp_voltages[] = { { "9", 0x10 },    { "9.5", 0x20 },
+			     { "10", 0x30 },   { "11", 0x40 },
+			     { "11.5", 0x50 }, { "12", 0x00 },
+			     { "12.5", 0x60 }, { "13", 0x70 },
+			     { "13.5", 0x80 }, { "14", 0x90 },
+			     { "14.5", 0xa0 }, { "15.5", 0xb0 },
+			     { "16", 0xc0 },   { "16.5", 0xd0 },
+			     { "17", 0xe0 },   { "18", 0xf0 },
+			     { "21", 0xf2 },   { "25", 0xf1 },
+			     { NULL, 0x00 } },
+  logic_voltages[] = { { "5", 0x00 },
+		       { "3.3", 0x01 },
+		       { "2.5", 0x02 },
+		       { "1.8", 0x03 },
+		       { NULL, 0x00 } },
+  t48_custom_vcc_voltages[] = { { "1.75", 0x01 },
                          { "1.8",  0x02 },
                          { "1.9",  0x03 },
                          { "2",  0x04 },
@@ -122,7 +140,7 @@ static struct voltage_s {
                          { "6.8",  0x3e },
                          { "6.9",  0x3f },
                          { NULL, 0x00 } },
-  t48_vpp_voltages[] = { { "9", 0x00 },
+  t48_custom_vpp_voltages[] = { { "9", 0x00 },
                          { "9.5", 0x01 },
                          { "10", 0x03 },
                          { "11", 0x07 },
@@ -140,19 +158,24 @@ static struct voltage_s {
                          { "18", 0x23 },
                          { "21", 0x2f },
                          { "25", 0x3e },
-                         { NULL, 0x00 } },
-  logic_voltages[] = { { "3.5", 0x04 },
-		       { "3.3", 0x03 },
-		       { "3", 0x02 },
-		       { "2.5", 0x01 },
-		       { "2.35", 0x00 },
-		       { NULL, 0x00 } };
+                         { NULL, 0x00 } };
+
+static struct spi_speed_s {
+	const char *name;
+	uint8_t value;
+} t48_spi_speeds[] = {
+			{ "3", 0x0 }, { "3.0", 0x00 },
+			{ "7.5", 0x01 },
+			{ "15", 0x02 }, { "15.0", 0x02 },
+			{ "30", 0x03 }, { "30.0", 0x03 },
+			{ NULL, 0x00 } };
 
 static struct option long_options[] = {
 	{ "pulse", required_argument, NULL, 2 },
 	{ "vpp", required_argument, NULL, 2 },
 	{ "vdd", required_argument, NULL, 2 },
 	{ "vcc", required_argument, NULL, 2 },
+	{ "speed", required_argument, NULL, 2},
 	{ "fuses", no_argument, NULL, 1 },
 	{ "uid", no_argument, NULL, 1 },
 	{ "lock", no_argument, NULL, 1 },
@@ -270,7 +293,7 @@ int get_programmer_version(uint8_t *version)
 	return EXIT_SUCCESS;
 }
 
-void print_supported_programmers_and_exit()
+void print_supported_programmers_and_exit(cmdopts_t *cmdopts)
 {
 	fprintf(stderr, "tl866a:  TL866CS/A\n"
 			"tl866ii: TL866II+\n"
@@ -280,7 +303,7 @@ void print_supported_programmers_and_exit()
 	exit(EXIT_SUCCESS);
 }
 
-void print_connected_programmer_and_exit()
+void print_connected_programmer_and_exit(cmdopts_t *cmdopts)
 {
 	minipro_handle_t *handle = minipro_open(NO_VERBOSE);
 	if (!handle) {
@@ -368,6 +391,11 @@ static struct voltage_s *get_vpp_voltages(minipro_handle_t *handle)
 		return tl866ii_vpp_voltages;
 		break;
 	case MP_T48:
+		if (handle->device->flags.custom_protocol)
+			return t48_custom_vpp_voltages;
+		else
+			return t48_vpp_voltages;
+		break;
 	case MP_T56:
 		return t48_vpp_voltages;
 		break;
@@ -386,6 +414,11 @@ static struct voltage_s *get_vcc_voltages(minipro_handle_t *handle)
 		return tl866ii_vcc_voltages;
 		break;
 	case MP_T48:
+		if (handle->device->flags.custom_protocol)
+			return t48_custom_vcc_voltages;
+		else
+			return t48_vcc_voltages;
+		break;
 	case MP_T56:
 		return t48_vcc_voltages;
 		break;
@@ -438,6 +471,19 @@ int set_voltage(minipro_handle_t *handle, char *value, uint8_t *target,
 	return EXIT_FAILURE;
 }
 
+int set_spi_speed(minipro_handle_t *handle, char *value, uint8_t *spi_speed) 
+{
+	struct spi_speed_s *speed = t48_spi_speeds;
+	while (speed->name) {
+		if (!strcasecmp(speed->name, value)) {
+			*spi_speed = speed->value;
+			return EXIT_SUCCESS;
+		}
+		speed++;
+	}
+	return EXIT_FAILURE;
+}
+
 int get_device(minipro_handle_t *handle)
 {
 	db_data_t db_data;
@@ -641,6 +687,11 @@ int parse_options(minipro_handle_t *handle, int argc, char **argv)
 						VCC_VOLTAGE))
 					return EXIT_FAILURE;
 				break;
+			case 4:
+				if (set_spi_speed(handle, optarg,
+						&handle->cmdopts->spi_speed))
+					return EXIT_FAILURE; 
+				break;
 			default:
 				return EXIT_FAILURE;
 			}
@@ -672,6 +723,10 @@ int parse_options(minipro_handle_t *handle, int argc, char **argv)
 						&device->voltages.vcc,
 						VCC_VOLTAGE))
 					return EXIT_FAILURE;
+			} else if (!strcasecmp(option, "speed")) {
+				if (set_spi_speed(handle, value,
+						&handle->cmdopts->spi_speed))
+					return EXIT_FAILURE;
 			} else
 				return EXIT_FAILURE;
 			break;
@@ -811,15 +866,15 @@ void parse_cmdline(int argc, char **argv, cmdopts_t *cmdopts)
 	void (*p_func)(cmdopts_t *) = NULL;
 
 	memset(cmdopts, 0, sizeof(cmdopts_t));
-	long_options[4].flag = &cmdopts->filter_fuses;
-	long_options[5].flag = &cmdopts->filter_uid;
-	long_options[6].flag = &cmdopts->filter_locks;
+	long_options[5].flag = &cmdopts->filter_fuses;
+	long_options[6].flag = &cmdopts->filter_uid;
+	long_options[7].flag = &cmdopts->filter_locks;
 	while ((c = getopt_long(argc, argv,
 				"lL:q:Qkd:ea:zEbTuPvxyr:w:m:p:c:o:iIsSVhDtf:F:",
 				long_options, NULL)) != -1) {
 		switch (c) {
 		case 0:
-		case 2: /* Skip vdd, vcc, vpp, pulse here */
+		case 2: /* Skip vdd, vcc, vpp, pulse, speed here */
 			break;
 		case 3:
 			cmdopts->infoic_path = optarg; /* Custom infoic.xml */
