@@ -280,6 +280,9 @@ int t48_begin_transaction(minipro_handle_t *handle)
 		if (device->voltages.raw_voltages & 0x80000000)
 			msg[22] = (device->voltages.raw_voltages >> 16) & 0x0f;
 
+		msg[24] = 1;    // need to set it to 1 in order for byte[28] to take the desired effect 
+		msg[28] = handle->cmdopts->spi_speed;
+
 		format_int(&(msg[40]),
 			   handle->device->package_details.packed_package, 4,
 			   MP_LITTLE_ENDIAN);
@@ -1026,6 +1029,7 @@ int t48_firmware_update(minipro_handle_t *handle, const char *firmware)
 
 static pin_driver_t last_pinstate[T48_NPINS];
 static uint8_t last_direction[T48_NPINS];
+static uint8_t last_output[T48_NPINS];
 static int last_vcc=0;
 static int last_j1vcc=1;
 
@@ -1250,6 +1254,13 @@ int t48_set_vccio_voltagef(minipro_handle_t *handle,float vccio,float tolerance)
 int t48_reset_state(minipro_handle_t *handle)
 {
 	uint8_t msg[48];
+	memset(last_direction, 0, sizeof(last_direction));
+	/* Set last_output to invalid value so we always change it if not set before */
+	memset(last_output, 0xff, sizeof(last_direction));
+	/* Also clear VCC/GND/VPP values */
+	memset(last_pinstate, 0, sizeof(last_pinstate));
+	last_vcc = 0;
+	last_j1vcc = 1;
 	memset(msg, 0, sizeof(msg));
 	/* Reset pin drivers state */
 	msg[0] = T48_RESET_PIN_DRIVERS;
@@ -1262,6 +1273,9 @@ int t48_set_input_and_pullup(minipro_handle_t *handle)
   /* Sets all pins to input and with pull down active
    */
 	uint8_t msg[48];
+	memset(last_direction, 0, sizeof(last_direction));
+	/* Set last_output to invalid value so we always change it if not set before */
+	memset(last_output, 0xff, sizeof(last_direction));
 	memset(msg, 0, sizeof(msg));
 	/* Reset pin drivers state */
 	msg[0] = T48_SET_PULLUPS;
@@ -1273,6 +1287,8 @@ int t48_set_input_and_pulldown(minipro_handle_t *handle)
   /* Sets all pins to input and with pull up active
    */
 	uint8_t msg[48];
+	memset(last_direction, 0, sizeof(last_direction));
+	/* Set last_output to invalid value so we always change it if not set before */
 	memset(msg, 0, sizeof(msg));
 	/* Reset pin drivers state */
 	msg[0] = T48_SET_PULLDOWNS;
@@ -1316,12 +1332,13 @@ int t48_set_zif_state(minipro_handle_t *handle, uint8_t *zif)
 	uint8_t msg[8];
 	memset(msg, 0, sizeof(msg));
 	for (int i = 0; i < T48_NPINS; i++) {
-          if(last_direction[i]==MP_PIN_DIRECTION_OUT) {
+          if(last_direction[i]==MP_PIN_DIRECTION_OUT && last_output[i] != zif[i]) {
             msg[0] = T48_SET_OUT;
             msg[1] = zif[i];
             msg[4] = i;
             if (msg_send(handle->usb_handle, msg, 8))
               return EXIT_FAILURE;
+	    last_output[i] = zif[i];
           }
 	}
 
@@ -1370,11 +1387,28 @@ int t48_screen_voltages(minipro_handle_t *handle) {
 	return EXIT_SUCCESS;
 }
 
+
+/*
+  VPP[0..15]
+  9.0,  9.5, 10.0, 11.0, 11.5, 12.0, 12.5, 13.0,
+  13.5, 14.0, 14.5, 15.5, 16.0, 16.5, 17.0, 18.0
+
+  VCC[0..15]
+  1.9, 2.7, 3.0, 3.3, 3.6, 3.9, 4.1, 4.5,
+  4.8, 5.0, 5.3, 5.5, 6.0, 6.3, 6.5, 7.0
+ */
+
 int t48_set_voltages(minipro_handle_t *handle, uint8_t vcc, uint8_t vpp)
 {
-	int err=t48_set_vcc_voltage(handle,vcc);
+	/* Internal VCC table firmware map */
+	static const uint8_t vcc_t[] = { 3, 13, 16, 19, 23, 27, 29, 34,
+					 38, 40, 44, 47, 52, 56, 59, 62};
+	int err=t48_set_vcc_voltage(handle,vcc_t[vcc]);
 	if(err) return err;
-	return t48_set_vpp_voltage(handle,vpp);
+	/* Internal VPP table firmware map */
+	static const uint8_t vpp_t[] = {0, 1, 3, 6, 9, 10, 13, 14, 17, 18,
+					20, 24, 26, 28, 30, 35};
+	return t48_set_vpp_voltage(handle,vpp_t[vpp]);
 }
 
 int t48_set_pin_drivers(minipro_handle_t *handle, pin_driver_t *pins)
