@@ -77,6 +77,8 @@
 #define TL866A_BOOTLOADER_SIZE	   0x1800
 #define TL866A_FIRMWARE_BLOCK_SIZE 0x50
 
+#define TL866A_NPINS		   40
+
 typedef struct zif_pins_s {
 	uint8_t pin;
 	uint8_t latch;
@@ -594,7 +596,7 @@ int tl866a_hardware_check(minipro_handle_t *handle)
 	uint8_t msg[64];
 	memset(msg, 0, sizeof(msg));
 
-	uint8_t i, errors = 0;
+	uint8_t i, j, errors = 0;
 	/* Reset pin drivers state */
 	msg[0] = TL866A_RESET_PIN_DRIVERS;
 	if (msg_send(handle->usb_handle, msg, 10)) {
@@ -708,7 +710,7 @@ int tl866a_hardware_check(minipro_handle_t *handle)
 		}
 		msg[0] = TL866A_SET_LATCH;
 		msg[1] = 0;
-		msg[7] = 1;
+		msg[7] = gnd_pins[i].latch == 9 ? 9 : 1;
 		msg[8] = gnd_pins[i].oe;
 		msg[9] = gnd_pins[i].latch;
 		msg[10] = gnd_pins[i].mask;
@@ -745,6 +747,68 @@ int tl866a_hardware_check(minipro_handle_t *handle)
 			read_buffer[6 + gnd_pins[i].pin] ? "Bad" : "OK");
 	}
 
+	fprintf(stderr, "\n");
+	/* Testing logic state */
+
+	/* Reset and set pullups */
+	memset(msg, 0, sizeof(msg));
+	msg[0] = TL866A_RESET_PIN_DRIVERS;
+	msg[1] = 1;
+	if (msg_send(handle->usb_handle, msg, 10)) {
+		return EXIT_FAILURE;
+	}
+
+	for (i = 0; i < TL866A_NPINS; i++) {
+		uint8_t pin_offset = 7 + zif_table[i];
+		msg[0] = TL866A_SET_DIR;
+		/* Set all pins as input except pin under test */
+		memset(msg + 7, 1, TL866A_NPINS);
+		msg[pin_offset] = 0;
+		if (msg_send(handle->usb_handle, msg, 47)) {
+			return EXIT_FAILURE;
+		}
+		memset(msg, 0, sizeof(msg));
+		msg[0] = TL866A_SET_OUT;
+		if (msg_send(handle->usb_handle, msg, 47)) {
+			return EXIT_FAILURE;
+		}
+		msg[0] = TL866A_READ_ZIF_PINS;
+		if (msg_send(handle->usb_handle, msg, 18)) {
+			return EXIT_FAILURE;
+		}
+		if (msg_recv(handle->usb_handle, read_buffer,
+			     sizeof(read_buffer))) {
+			return EXIT_FAILURE;
+		}
+		if (read_buffer[1]) {
+			msg[0] = TL866A_RESET_PIN_DRIVERS;
+			if (msg_send(handle->usb_handle, msg, 10)) {
+				minipro_close(handle);
+				return EXIT_FAILURE;
+			}
+			if (minipro_end_transaction(handle)) {
+				return EXIT_FAILURE;
+			}
+			fprintf(stderr,
+				"Overcurrent protection detected while testing Logic level driver "
+				"%u!\007\n", i + 1);
+			return EXIT_FAILURE;
+		}
+		if (read_buffer[7 + i])
+			errors++;
+		fprintf(stderr, "Logic 0 pin %u is %s\n", i + 1,
+			read_buffer[7 + i] ? "Bad" : "OK");
+		for (j = 0; j < TL866A_NPINS; j++) {
+			if (i == j)
+				continue;
+			if (!read_buffer[7 + j]) {
+				fprintf(stderr, "Unexpected state of pin %u when testing logic pin %u: shorted?\n",
+					j + 1, i + 1);
+			errors++;
+			}
+		}
+	}
+
 	fprintf(stderr, "\n");
 	/* Testing VPP overcurrent protection */
 	msg[0] = TL866A_SET_LATCH;
@@ -1183,8 +1247,8 @@ static zif_pins_t *get_pwr_pin(int pin, zif_pins_t *table, int count)
 static int pwr_init(minipro_handle_t *handle, uint8_t *vector, size_t pin_count)
 {
 	assert(vector != NULL);
-	static uint8_t pwr[] = { TL866A_POWER_ON,
-				 0x00,
+	static uint8_t pwr[] = { TL866A_RESET_PIN_DRIVERS,
+				 0x01,
 				 0x00,
 				 0x00,
 				 0x00,
@@ -1205,20 +1269,12 @@ static int pwr_init(minipro_handle_t *handle, uint8_t *vector, size_t pin_count)
 				 0x07,
 				 0x00 };
 
-	/* This is a trick to switch on all pull-up resistors
-	 * First switch on VCC
-	 * The VCC will be applied to pin 40
-	 * Also this will enable all pull-up resistors
-	 */
-	if (msg_send(handle->usb_handle, pwr, 4)) {
+	/* Switch on all pullup resistors */
+	if (msg_send(handle->usb_handle, pwr, 10)) {
 		return EXIT_FAILURE;
 	}
 
-	/* Now disable all pin drivers */
 	pwr[0] = TL866A_SET_LATCH;
-	if (msg_send(handle->usb_handle, pwr, sizeof(pwr))) {
-		return EXIT_FAILURE;
-	}
 
 	/* Scan the test vector for V and G */
 	zif_pins_t *pwr_pin;
@@ -1230,6 +1286,7 @@ static int pwr_init(minipro_handle_t *handle, uint8_t *vector, size_t pin_count)
 					      sizeof(gnd_pins) /
 						      sizeof(gnd_pins[0]));
 			if (!pwr_pin) {
+				fprintf(stderr, "Pin %zu not supported for GND on TL866A/CS!\n", i + 1);
 				return EXIT_FAILURE;
 			}
 			pwr[(pwr_pin->latch - 2) * 2 + 10] |= pwr_pin->mask;
@@ -1238,6 +1295,7 @@ static int pwr_init(minipro_handle_t *handle, uint8_t *vector, size_t pin_count)
 					      sizeof(vcc_pins) /
 						      sizeof(vcc_pins[0]));
 			if (!pwr_pin) {
+				fprintf(stderr, "Pin %zu not supported for VCC on TL866A/CS!\n", i + 1);
 				return EXIT_FAILURE;
 			}
 			pwr[(pwr_pin->latch - 2) * 2 + 10] &= pwr_pin->mask;
@@ -1362,8 +1420,7 @@ int tl866a_logic_ic_test(minipro_handle_t *handle)
 	int ret = EXIT_FAILURE;
 
 	/* Check for invalid pin count */
-	if (handle->device->package_details.pin_count > 32 ||
-	    handle->device->package_details.pin_count % 2) {
+	if (handle->device->package_details.pin_count % 2) {
 		fprintf(stderr, "Invalid pin count!\n");
 		return ret;
 	}
@@ -1375,7 +1432,6 @@ int tl866a_logic_ic_test(minipro_handle_t *handle)
 	if (!(result = do_ic_test(handle))) {
 		fprintf(stderr, "Error running logic test.\n");
 	} else if (handle->cmdopts->logicic_out) {
-		/* WARNING: untested */
 		ret = write_logic_file(handle, result, result);
 	} else {
 		int errors = 0, err;
